"""
AbstractMusic vendored dependencies.

This package contains **source code vendored** from upstream projects under
permissive licenses (MIT/Apache/BSD), so `abstractmusic` can remain fully
open-source and not rely on runtime "remote code" execution.
"""

